const relayKey = '123456';
import { connect } from 'cloudflare:sockets';

async function fetch(request, env, ctx) {
    if (request.headers.get("Upgrade") !== "websocket") {
        const { tlsClientAuth, tlsExportedAuthenticator, ...clientInfo } = request.cf;
        return new Response(JSON.stringify(clientInfo, null, 4));
    }

    if (request.headers.get("Relay-Key") !== relayKey) {
        return new Response(null, { status: 401 });
    }

    const relayTarget = request.headers.get("Relay-Target");
    if (!relayTarget) {
        return new Response(null, { status: 400 });
    }

    const [hostname, portString] = relayTarget.split(":");
    const [client, server] = Object.values(new WebSocketPair());
    const headers = { "Relay-Random": createRandomString(1, 1024) };
    if (hostname === "localhost") {
        return new Response(null, { status: 101, webSocket: client, headers });
    }

    try {
        const remoteSocket = connect({ hostname, port: Number(portString) });
        await remoteSocket.opened;

        server.accept();
        const webSocket = createWebSocket(server);

        const promise1 = webSocket.readable.pipeTo(remoteSocket.writable);
        const promise2 = remoteSocket.readable.pipeTo(webSocket.writable);
        ctx.waitUntil(Promise.race([promise1, promise2]));

        return new Response(null, { status: 101, webSocket: client, headers });
    } catch (error) {
        return new Response(error, { status: 502, headers });
    }
}

function createWebSocket(webSocket) {
    const readable = new ReadableStream({
        start(controller) {
            webSocket.addEventListener('close', () => controller.close());
            webSocket.addEventListener('error', (error) => controller.error(error));
            webSocket.addEventListener('message', (message) => controller.enqueue(message.data));
        }
    });

    const writable = new WritableStream({
        write(data) {
            webSocket.send(data);
        },
        close() {
            webSocket.close();
        },
        abort() {
            webSocket.close();
        }
    });
    return { readable, writable };
}

function createRandomString(minLength, maxLength) {
    const length = Math.floor(Math.random() * (maxLength - minLength)) + minLength;
    const chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
    let result = '';
    for (let i = 0; i < length; i++) {
        result += chars[Math.floor(Math.random() * chars.length)];
    }
    return result;
}

export default { fetch }